# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::KTab;

# Katie::KTab->new() returns a KTab object with the following methods
# available:
# - begin_append()
# - append($mount_point, $user)
# - check_umount($mount_point)
# - begin_remove()
# - remove($mount_point)

use strict;
use Katie::Constants;

use Fcntl qw(LOCK_EX LOCK_UN);
use FileHandle;

my($ktab_fn) = RUN_DIR . "/ktab";
my($ktab_lock_fn) = "$ktab_fn.lock";

sub new {
  my($self) = {};

  $self->{username} = get_user_name();
  $self->{ktab_lock_f} = lock_ktab();

  return bless $self;
}

sub begin_append {
  my($self) = @_;
  my($ktab) = new FileHandle;

  if (!$ktab->open("$ktab_fn", ">>")) {
    die "Couldn't open $ktab_fn: $!\n";
  }

  $self->{ktab} = $ktab;
}

sub append {
  my($self, $mount_point) = @_;

  $self->{ktab}->print("$mount_point\t$self->{username}\n");
}

sub check_umount {
  my($self, $mount_point) = @_;
  my($ktab) = new FileHandle;
  local($_);

  if (!$ktab->open($ktab_fn, "<")) {
    die "Couldn't open $ktab_fn: $!\n";
  }

  while (<$ktab>) {
    if (/^(.*)\t(.*)$/) {
      return 1 if ($1 eq $mount_point and $2 eq $self->{username});
    }
  }
  return 0;
}

sub begin_remove {
  my($self) = @_;
  my($ktab) = new FileHandle;
  my($ktab_new) = new FileHandle;

  if (!$ktab->open($ktab_fn, "<")) {
    die "Couldn't open $ktab_fn: $!\n";
  }

  if (!$ktab_new->open("$ktab_fn.new", ">")) {
    die "Couldn't open $ktab_fn.new: $!\n";
  }

  $self->{ktab} = $ktab;
  $self->{ktab_new} = $ktab_new;
}

sub remove {
  my($self, $mount_point) = @_;
  local($_);

  while ($_ = $self->{ktab}->getline()) {
    if (/^(.*)\t(.*)$/) {
      last if ($1 eq $mount_point);
    } else {
      next;
    }
    $self->{ktab_new}->print($_);
  }
  while ($_ = $self->{ktab}->getline()) {
    $self->{ktab_new}->print($_);
  }

  delete($self->{ktab});
  delete($self->{ktab_new});

  if (!rename("$ktab_fn.new", $ktab_fn)) {
    die "$0: couldn't rename $ktab_fn.new to $ktab_fn: $!\n";
  }
}

# return the name of the calling user
sub get_user_name {
  my(@pw) = getpwuid($<);

  die "$0: who are you?\n" if (!@pw);

  if ($pw[0] =~ /\t/) {
    die "$0: why does your username have a TAB in it?\n";
  }

  return $pw[0];
}

# lock the ktab file for the duration of the program
sub lock_ktab {
  my($ktab_lock_f) = new FileHandle;

  if (!$ktab_lock_f->open($ktab_lock_fn, ">")) {
    die "$0: couldn't open $ktab_lock_fn: $!\n";
  }

  if (!flock($ktab_lock_f, LOCK_EX)) {
    die "$0: couldn't lock $ktab_lock_fn: $!\n";
  }

  return $ktab_lock_f;
}

1;
