# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::VOBClear;

use strict;
use Katie::NFSerrno;

# start closing filehandles when this number is exceeded
use constant HI_NUM_FH_CACHE => 10;

# keep closing them until this number is reached
use constant LO_NUM_FH_CACHE => 5;

# This package implements the VOB cleartext cache.
#
# new($store_dir_c, $store_obj) returns a Katie::VOBClear object.
#
# $store_dir_c: directory to cache VOB cleartext in
# $store_obj: a Katie::VOBStore object to retrieve cleartext from
#
# methods:
#   read

sub new {
  shift(@_);
  my($store_dir_c, $store_obj) = @_;
  my($self) = {
    store_dir_c => $store_dir_c,  # cleartext storage directory
    store_obj   => $store_obj,    # Katie::VOBStore object
    fh_cache    => { },           # cache for filehandles
    fh_num      => 0,             # number of filehandles in fh_cache
  };

  return bless $self;
}

# read data from the given VOB file
sub read {
  my($self, $element_id, $inode_id, $offset, $count) = @_;
  my($fh) = new FileHandle;
  my($data, $data_len);

  # check if we already have the file open
  if (!$self->{fh_cache}->{$inode_id}) {

    # file is not open -- we will be using one more fh, so check if we
    # need to cut the number down
    if ($self->{fh_num} >= HI_NUM_FH_CACHE) {
      $self->close_filehandles();
    }

    # attempt to open the file
    my($fn) = "$self->{store_dir_c}/$inode_id";
    if (!$fh->open($fn, "<")) {

      # can't open the file -- attempt to construct it
      $self->{store_obj}->construct_revision($element_id, $inode_id, $fn);
      if (!$fh->open($fn, "<")) {

        # construct failed -- return error
        warn "vob_read: open($fn): $!\n";
        return (undef, NFSERR_IO);
      }
    }

    # create the cache entry and bump fh_num
    $self->{fh_cache}->{$inode_id} = {
      fh    => $fh,
      atime => time(),
    };
    $self->{fh_num}++;

  # file is already open -- set $fh and update the atime
  } else {
    my($fh_cache) = $self->{fh_cache}->{$inode_id};
    $fh = $fh_cache->{fh};
    $fh_cache->{atime} = time();
  }

  sysseek($fh, $offset, 0);
  $data_len = sysread($fh, $data, $count);
  if (!defined($data_len)) {
    return (undef, NFSERR_IO);
  }
  return ($data_len, $data);
}

sub close_filehandles {
  my($self) = @_;
  my(%fh_by_time);
  my($num_to_close) = HI_NUM_FH_CACHE - LO_NUM_FH_CACHE;

  # sort the list in fh_cache by atime so we close the oldest first
  for my $inode_id (
        sort { $self->{fh_cache}->{$a}->{atime} <=>
               $self->{fh_cache}->{$b}->{atime}
             }
        keys %{$self->{fh_cache}}
      ) {

    # quit if there are no more to close
    if ($num_to_close-- <= 0) {
      return;
    }

    # close the fh
    $self->{fh_cache}->{$inode_id}->{fh}->close();
    delete($self->{fh_cache}->{$inode_id});
    $self->{fh_num}--;
  }
}

1;
