# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::ViewFile;

use strict;
use Katie::NFSerrno;
use Katie::VOB;
use Katie::View;
use Katie::Constants;

require Exporter;

@Katie::ViewFile::ISA = qw(Exporter);

@Katie::ViewFile::EXPORT = qw(
  view_private_get_attr
  view_private_create_file
  view_private_settime
  view_private_setmodes
  view_private_setsize
  view_private_unlink
  view_private_read
  view_private_write
);

# XXX ugh
my($NFREG)  = 1; # regular file
my($NFDIR)  = 2; # directory

sub dir_mtime_get {
  return Katie::View::dir_mtime_get(@_);
}

# return the attributes associated with the given view-private
# file
sub view_private_get_attr_file {
  my($view_id, $inode_id, $inode) = @_;
  my($view_store) = $view_cache{$view_id}->{store};
  my(@stat);

  @stat = lstat("$view_store/$inode_id");
  if (!@stat) {
    warn "view_private_get_attr_file: couldn't stat $view_store/$inode_id\n";
    return NFSERR_IO;
  }

  return (
    $NFREG,               # type
    $inode->{mode},       # mode
    $inode->{link_count}, # nlink
    $inode->{uid},        # uid
    $inode->{gid},        # gid
    $stat[7],             # size
    $stat[11],            # blocksize
    0,                    # rdev
    $stat[12],            # blocks
    $view_id,             # fsid
    $inode_id,            # fileid
    $stat[8],             # atime
    $stat[9],             # mtime
    $stat[10]             # ctime
  );
}

# return the attributes associated with the given view-private
# directory
sub view_private_get_attr_dir {
  my($view_id, $inode_id, $inode) = @_;
  my($mtime) = dir_mtime_get($view_id, $inode_id);

  if (!defined($mtime)) {
    warn "view_private_get_attr_dir: " .
         "mtime not defined for $view_id/$inode_id\n";
    $mtime = 0;
  }

  return (
    $NFDIR,               # type
    $inode->{mode},       # mode
    $inode->{link_count}, # nlink
    $inode->{uid},        # uid
    $inode->{gid},        # gid
    1024,                 # size
    1024,                 # blocksize
    0,                    # rdev
    1,                    # blocks
    $view_id,             # fsid
    $inode_id,            # fileid
    $mtime,               # atime
    $mtime,               # mtime
    $mtime                # ctime
  );
}

# prereq: get_inode_type must have already been called on $inode_id
# return the attributes associated with the given view-private
# file/directory
sub view_private_get_attr {
  my($view_id, $inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($inode) = $inode_cache[$vob_id]->{$inode_id};
  my(@stat, $type);

  if ($inode->{type} == eREG) {
    return view_private_get_attr_file($view_id, $inode_id, $inode);
  } elsif ($inode->{type} == eDIR) {
    return view_private_get_attr_dir($view_id, $inode_id, $inode);
  } else {
    warn "view_private_get_attr: $view_id, $inode_id has unknown type\n";
    return NFSERR_IO;
  }
}

# read data from the given view-private file
sub view_private_read {
  my($view_id, $inode_id, $offset, $count) = @_;
  my($view_store) = $view_cache{$view_id}->{store};
  my($fh) = new FileHandle;
  my($data, $data_len);

  if (!$fh->open("$view_store/$inode_id", "<")) {
    warn "view_private_read: couldn't open $view_store/$inode_id: $!\n";
    return (undef, NFSERR_NOENT);
  }
  sysseek($fh, $offset, 0);
  $data_len = sysread($fh, $data, $count);
  if (!defined($data_len)) {
    warn "view_private_read: error reading $view_store/$inode_id: $!\n";
    return (undef, NFSERR_IO);
  }
  return ($data_len, $data);
}

# write data to the given view-private file
sub view_private_write {
  my($view_id, $inode_id, $offset, $data) = @_;
  my($view_store) = $view_cache{$view_id}->{store};
  my($fh) = new FileHandle;
  my($data_len);

  if (!$fh->open("$view_store/$inode_id", "+<")) {
    warn "view_private_write: couldn't open $view_store/$inode_id: $!\n";
    return NFSERR_NOENT;
  }
  sysseek($fh, $offset, 0);
  $data_len = syswrite($fh, $data);
  if (!defined($data_len)) {
    warn "view_private_write: error writing $view_store/$inode_id: $!\n";
    return NFSERR_IO;
  }
  if ($data_len != length($data)) {
    warn "view_private_write: only wrote $data_len bytes of " .
         length($data) . " to $view_store/$inode_id\n";
    return NFSERR_IO;
  }
  return 0;
}

# set the attributes associated with the given view-private
# file/directory
sub view_private_setmodes {
  my($view_id, $inode_id, $mode, $uid, $gid) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my(%update);

  if ($mode != -1) {
    $mode &= 07777;
    $update{mode} = $mode;
    $inode_cache[$vob_id]->{$inode_id}->{mode} = $mode;
  }
  if ($uid != -1) {
    $update{uid} = $uid;
    $inode_cache[$vob_id]->{$inode_id}->{uid} = $uid;
  }
  if ($gid != -1) {
    $update{gid} = $gid;
    $inode_cache[$vob_id]->{$inode_id}->{gid} = $gid;
  }

  $vob_db->update("view_private", \%update, "id = $inode_id");

  return 1;
}

# XXX to be replaced with some real view manager stuff
# set the atime/time associated with the given view-private
# file/directory
sub view_private_settime {
  my($view_id, $inode_id, $atime, $mtime) = @_;
  my($view_store) = $view_cache{$view_id}->{store};
  my(@stat);

  if ($atime == -1 || $mtime == -1) {
    @stat = lstat("$view_store/$inode_id");
    if ($atime == -1) {
      $atime = $stat[8];
    }
    if ($mtime == -1) {
      $mtime = $stat[9];
    }
  }

  if (!utime($atime, $mtime, "$view_store/$inode_id")) {
    warn "view_private_settime: utime($view_store/$inode_id): $!\n";
    return 0;
  }

  return 1;
}

# XXX to be replaced with some real view manager stuff
# set the size of the given view-private file/directory
sub view_private_setsize {
  my($view_id, $inode_id, $size) = @_;
  my($view_store) = $view_cache{$view_id}->{store};

  if (!truncate("$view_store/$inode_id", $size)) {
    warn "view_private_setsize: truncate($view_store/$inode_id): $!\n";
    return 0;
  }

  return 1;
}

# XXX to be replaced with some real view manager stuff
# create the given view-private regular file
sub view_private_create_file {
  my($view_id, $inode_id) = @_;
  my($view_store) = $view_cache{$view_id}->{store};
  my($fh) = new FileHandle;

  if (!$fh->open("$view_store/$inode_id", ">")) {
    warn "view_private_create_file: create($view_store/$inode_id): $!\n";
    return NFSERR_IO;
  }
}

sub view_private_unlink {
  my($view_id, $inode_id) = @_;
  my($view_store) = $view_cache{$view_id}->{store};

  if (!unlink("$view_store/$inode_id")) {
    # XXX log this error?  return 0?
    return NFSERR_IO;
  }
}

1;
