# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::ViewLookup;

use strict;
use Katie::VOB;
use Katie::View;
use Katie::ViewReaddir;
use Katie::ConfigSpec;
use Katie::Constants;

require Exporter;

@Katie::ViewLookup::ISA = qw(Exporter);

@Katie::ViewLookup::EXPORT = qw(
  lookup_name
  lookup_name_view
  lookup_name_vupr
);

use strict;

# given a directory of type dVIEW, look for the given @@ file (the
# $name must already be SQL-escaped, and be passed without @@ on the
# end) and return the inode number.
# $dir_revision_type must be either iREVI or iVUPR
sub lookup_name_view_atat {
  my($view_id, $vob_id, $vob_db,
     $dir_revision_id, $dir_revision_type, $name) = @_;
  my($row);

  if ($dir_revision_type == iREVI) {
    $row = $vob_db->select("
        SELECT element.id2
        FROM   element, dir_entry
        WHERE  element.id = dir_entry.element_id
           AND dir_entry.dir_id = $dir_revision_id
           AND dir_entry.file_name = '$name';
    ")->fetchrow();
  } else {
    # get the view private directory ID from the checkout object
    $dir_revision_id =
        $inode_cache[$vob_id]->{$dir_revision_id}->{view_private_id};
    $row = $vob_db->select("
        SELECT element.id2
        FROM   element, dir_private_entry
        WHERE  element.id = dir_private_entry.file_id
           AND dir_private_entry.view_id = $view_id
           AND dir_private_entry.dir_id = $dir_revision_id
           AND dir_private_entry.file_name = '$name';
    ")->fetchrow();
  }

  return undef unless $row;

  if (get_inode_type($view_id, $row->{id2}) != iATAT) {
    warn "lookup_name_view_atat(@_): " .
         "get_inode_type($view_id, $row->{id2}) != iATAT";
    return undef;
  }

  return $row->{id2};
}

# given a directory of type dVIEW, look up the inode number
# associated with the given name
# postcond (not yet, but should be): get_inode_type() will have been
# called on the return value
sub lookup_name_view {
  my($view_id, $dir_inode_id, $dir_dpath, $name, $want_cout) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($dir_revision_id) = select_revision_id($view_id, $dir_inode_id,
                                            $dir_dpath);
  my($dir_revision_type);
  my($view_private_row);
  my($vob_row);
  my($atat);

  if (!$dir_revision_id) {
    warn "lookup_name_view: " .
         "select_revision_id($view_id, $dir_inode_id) returned undef\n";
    return undef;
  }

  $dir_revision_type = get_inode_type($view_id, $dir_revision_id);

  $atat = ($name =~ s/\@\@$//);
  $name = $vob_db->escape_sql($name);

  if ($atat) {
    return lookup_name_view_atat($view_id, $vob_id, $vob_db,
                                 $dir_revision_id, $dir_revision_type, $name);
  }

  # Look for a view-private file
  $view_private_row = $vob_db->select("
      SELECT file_id
      FROM   dir_private_entry
      WHERE  view_id = $view_id
         AND dir_id = $dir_inode_id
         AND file_name = '$name';
  ")->fetchrow();

  if ($view_private_row and
      !get_inode_type($view_id, $view_private_row->{file_id})) {
    warn "lookup_name_view(@_): " .
         "get_inode_type($view_id, $view_private_row->{file_id}) failed";
    undef $view_private_row;
  }

  # Look for a VOB file in either dir_entry or dir_private_entry
  # depending on whether the directory is checked out or not.
  if ($dir_revision_type == iREVI) {
    $vob_row = $vob_db->select("
        SELECT element_id AS file_id
        FROM   dir_entry
        WHERE  dir_id = $dir_revision_id
           AND file_name = '$name';
    ")->fetchrow();
  } elsif ($dir_revision_type == iCOUT) {
    # get the view private directory ID from the checkout object
    $dir_revision_id =
        $inode_cache[$vob_id]->{$dir_revision_id}->{view_private_id};
    $vob_row = $vob_db->select("
        SELECT file_id
        FROM   dir_private_entry
        WHERE  view_id = $view_id
           AND dir_id = $dir_revision_id
           AND file_name = '$name';
    ")->fetchrow();
  } else {
    warn "lookup_name_view(@_): revision returned " .
         "by select_revision_id() ($dir_revision_type) not REVI or COUT\n";
  }

  # check for VOB object taking precedence
  if ($vob_row) {

    if (get_inode_type($view_id, $vob_row->{file_id}) != iELEM) {
      warn "directory $dir_revision_id " .
           "contains non-element $vob_row->{file_id}\n";
      return undef;
    }

    # checked out elements appear on top of view private files
    if (is_checked_out($view_id, $vob_row->{file_id})) {

      # unless the element is a regular file...
      # (unless $want_cout is set)
      if ($inode_cache[$vob_id]->{$vob_row->{file_id}}->{type} == eREG and
          !$want_cout) {

        # ...and the config spec selects the CHECKEDOUT revision of
        # the element
        my($selected_object) =
            select_revision_id($view_id, $vob_row->{file_id},
                               "$dir_dpath/name");

        if (get_inode_type($view_id, $selected_object) == iCOUT) {
          # in which case it doesn't appear
          undef $vob_row;
        }
      }

      return $vob_row->{file_id} if $vob_row;
    }
  }

  return $view_private_row->{file_id} if $view_private_row;
  return $vob_row->{file_id} if $vob_row;
  return undef;
}

# given a directory of type dATAT, return the inode number associated
# with the given name
sub lookup_name_atat {
  my($view_id, $dir_inode_id, $name) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($query, $row);

  if ($name eq '..') {
    # we know that get_dir_type has already been called for this $dir_inode_id
    return $inode_cache[$vob_id]->{$dir_inode_id}->{id};
  }

  $name = $vob_db->escape_sql($name);
  $query = $vob_db->select("
      SELECT element_branch.id
      FROM   element_branch, element, branch
      WHERE  element_branch.element_id = element.id
         AND element.id2 = $dir_inode_id
         AND element_branch.parent_revision_id IS NULL
         AND element_branch.branch_id = branch.id
         AND branch.name = '$name';
  ");
  $row = $query->fetchrow();
  if ($row) {
    return $row->{id};
  }

  $query = $vob_db->select("
      SELECT revision.id
      FROM   label, label_type, revision, element, element_branch
      WHERE  label.label_type_id = label_type.id
         AND NOT label_type.pbranch
         AND label.revision_id = revision.id
         AND revision.element_branch_id = element_branch.id
         AND element_branch.element_id = element.id
         AND element.id2 = $dir_inode_id
         AND label_type.name = '$name';
  ");
  $row = $query->fetchrow();
  if ($row) {
    return $row->{id};
  }

  return undef;
}

# given a directory of type dELBR, return the inode number associated
# with the given name
sub lookup_name_elbr {
  my($view_id, $dir_inode_id, $name) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($query, $row);

  if ($name eq '..') {
    # we know that get_dir_type has already been called for this
    # $dir_inode_id
    # XXX what about $element_id?  $parent_revision_id?
    my($dir_inode) = $inode_cache[$vob_id]->{$dir_inode_id};

    if (!defined($dir_inode->{parent_revision_id})) {
      my($element_id) = $dir_inode->{element_id};
      return $inode_cache[$vob_id]->{$element_id}->{id2};
    } else {
      my($parent_revision_id) = $dir_inode->{parent_revision_id};
      return
          $inode_cache[$vob_id]->{$parent_revision_id}->{element_revision_id};
    }
  }

  if ($name eq 'LATEST') {
    $query = $vob_db->select("
        SELECT id
        FROM   revision
        WHERE  element_branch_id = $dir_inode_id
           AND num = (
                   SELECT max(num)
                   FROM   revision
                   WHERE  element_branch_id = $dir_inode_id
               );
    ");
    $row = $query->fetchrow();
    return $row->{id} if ($row);
    return undef;
  }

  if ($name =~ /^CHECKEDOUT(\.([0-9]+))?$/) {
    my($cout_id) = $2;
    my($checkout_query);
    if ($cout_id) {
      $checkout_query = "checkout.id = $cout_id";
    } else {
      $checkout_query = "checkout.view_id = $view_id";
    }
    $row = $vob_db->select("
        SELECT checkout.id
        FROM   checkout, revision
        WHERE  $checkout_query
           AND checkout.revision_id = revision.id
           AND revision.element_branch_id = $dir_inode_id;
    ")->fetchrow();
    return $row->{id} if ($row);
    return undef;
  }

  # Look for a revision number
  if ($name =~ /^[0-9]+$/) {
    $query = $vob_db->select("
        SELECT id
        FROM   revision
        WHERE  element_branch_id = $dir_inode_id
           AND num = $name;
    ");
    $row = $query->fetchrow();
    return $row->{id} if ($row);
    return undef;
  }

  $name = $vob_db->escape_sql($name);

  # Look for a label
  $row = $vob_db->select("
      SELECT revision.id
      FROM   revision, label, label_type
      WHERE  label.label_type_id = label_type.id
         AND label.revision_id = revision.id
         AND revision.element_branch_id = $dir_inode_id
         AND label_type.name = '$name';
  ")->fetchrow();
  return $row->{id} if ($row);

  # Branches
  $row = $vob_db->select("
      SELECT element_branch.id
      FROM   element_branch, branch, revision
      WHERE  element_branch.branch_id = branch.id
         AND element_branch.parent_revision_id = revision.id
         AND revision.element_branch_id = $dir_inode_id
         AND branch.name = '$name';
  ")->fetchrow();
  return $row->{id} if ($row);

  return undef;
}

# given a directory of type dREVI, return the inode number associated
# with the given name
sub lookup_name_revi {
  my($view_id, $dir_inode_id, $name) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($query, $row);

  if ($name eq '..') {
    # we know that get_dir_type has already been called for this
    # $dir_inode_id
    return $inode_cache[$vob_id]->{$dir_inode_id}->{element_branch_id};
  }

  # strip '@@' if it was supplied
  $name =~ s/\@\@$//;

  $name = $vob_db->escape_sql($name);
  $query = $vob_db->select("
      SELECT element_id
      FROM   dir_entry
      WHERE  dir_id = $dir_inode_id
         AND file_name = '$name';
  ");
  $row = $query->fetchrow();
  if (!$row) {
    return undef;
  }

  if (get_inode_type($view_id, $row->{element_id}) != iELEM) {
    warn "lookup_name_revi(@_): " .
         "get_inode_type($view_id, $row->{element_id}) != iELEM";
    return undef;
  }
  return $inode_cache[$vob_id]->{$row->{element_id}}->{id2};
}

# given a directory of type dCOUT, return the inode number associated
# with the given name
sub lookup_name_cout {
  my($view_id, $dir_inode_id, $name) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($vupr_dir_inode_id);
  my($query, $row);

  $vupr_dir_inode_id =
      $inode_cache[$vob_id]->{$dir_inode_id}->{view_private_id};

  # XXX should be able to figure this out w/o a db query
  $row = $vob_db->select("
      SELECT revision.element_branch_id
      FROM   revision, checkout
      WHERE  revision.id = checkout.revision_id
         AND checkout.id = $dir_inode_id;
  ")->fetchrow();
  if (!$row) {
    warn "couldn't get element_branch_id for $dir_inode_id\n";
    return undef;
  }
  my($element_branch_id) = $row->{element_branch_id};

  $name = $vob_db->escape_sql($name);
  $row = $vob_db->select("
      SELECT file_id
      FROM   dir_private_entry
      WHERE  dir_id = $vupr_dir_inode_id
         AND file_name = '$name';
  ")->fetchrow();

  return undef unless $row;

  return $inode_cache[$vob_id]->{$row->{file_id}}->{id2};
}

# given a directory of type dVUPR, return the inode number associated
# with the given name
sub lookup_name_vupr {
  my($view_id, $dir_inode_id, $name) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($row);

  $name = $vob_db->escape_sql($name);
  $row = $vob_db->select("
      SELECT file_id
      FROM   dir_private_entry
      WHERE  dir_id = $dir_inode_id
         AND file_name = '$name';
  ")->fetchrow();

  return undef unless $row;

  return $row->{file_id};
}

# given a directory of any type, return the inode number associated
# with the given name
sub lookup_name {
  my($view_id, $dir_inode_id, $dir_dpath, $name, $want_cout) = @_;
  my($dir_type);

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if (!$dir_type) {
    warn "lookup_name($view_id, $dir_inode_id): not a directory\n";
    return undef;
  }

  return $dir_inode_id if ($name eq '.');

  if ($dir_type == dVIEW) {
    return lookup_name_view($view_id, $dir_inode_id, $dir_dpath, $name,
                            $want_cout);
  } elsif ($dir_type == dATAT) {
    return lookup_name_atat($view_id, $dir_inode_id, $name);
  } elsif ($dir_type == dELBR) {
    return lookup_name_elbr($view_id, $dir_inode_id, $name);
  } elsif ($dir_type == dREVI) {
    return lookup_name_revi($view_id, $dir_inode_id, $name);
  } elsif ($dir_type == dVUPR) {
    return lookup_name_vupr($view_id, $dir_inode_id, $name);
  } elsif ($dir_type == dCOUT) {
    return lookup_name_cout($view_id, $dir_inode_id, $name);
  }
  warn "lookup_name($view_id, $dir_inode_id, $name): dir_type is $dir_type\n";
  return undef;
}

1;
