# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::ViewReaddir;

use strict;
use Katie::NFSerrno;
use Katie::VOB;
use Katie::View;
use Katie::ConfigSpec;
use Katie::Constants;

require Exporter;

@Katie::ViewReaddir::ISA = qw(Exporter);

@Katie::ViewReaddir::EXPORT = qw(
  get_dir_type
  view_readdir
);

# given a inode_id, return undef if it isn't a directory, or one of
# the following if it is:
#   dVIEW
#   dATAT
#   dELBR
#   dREVI
#   dVUPR
#   dCOUT

sub get_dir_type {
  my($view_id, $dir_inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($inode_type);

  # Check for bad inode
  if (!($inode_type = get_inode_type($view_id, $dir_inode_id))) {
    return undef;
  }

  # Get the element ID of a revision so we can get the real file type
  if ($inode_type == iREVI or $inode_type == iCOUT) {
    if (!($dir_inode_id = get_element_id($view_id, $dir_inode_id))) {
      return undef;
    }
  }

  # Check for dVIEW, dVUPR, dREVI, or dCOUT -- these should be directories
  if ($inode_type == iELEM or $inode_type == iVUPR or
      $inode_type == iREVI or $inode_type == iCOUT) {
    if ($inode_cache[$vob_id]->{$dir_inode_id}->{type} == eDIR) {
      return $inode_type;
    } else {
      # Element isn't a directory
      return undef;
    }

  # Check for dATAT, dELBR -- these are always directories
  } elsif ($inode_type == iATAT or $inode_type == iELBR) {
    return $inode_type;

  # Object is not something that can be seen in the view
  } else {
    warn "get_dir_type(@_): bad inode_type $inode_type\n";
    return undef;
  }
}

# given a directory of type dVIEW, return a listing of the files in
# the directory
sub readdir_view {
  my($view_id, $dir_inode_id, $dir_dpath) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($dir_revision_id) = select_revision_id($view_id, $dir_inode_id,
                                            $dir_dpath);
  my(%view_list);
  my(%vob_list);
  my($query, $row);

  if (!$dir_revision_id) {
    warn "readdir_view: " .
         "select_revision_id(@_) returned undef\n";
    return (NFSERR_IO);
  }

  # put the list of view-private entries in the directory into %view_list
  $query = $vob_db->select("
      SELECT id AS dir_entry_id, file_id, file_name
      FROM   dir_private_entry
      WHERE  view_id = $view_id
         AND dir_id = $dir_inode_id;
  ");
  $view_list{$row->{file_name}} = $row while $row = $query->fetchrow();

  my($dir_revision_type) = get_inode_type($view_id, $dir_revision_id);

  # TODO: call get_vob_directory_list() instead
  # Get the contents of the directory from either dir_entry or
  # dir_private_entry depending on whether the directory is checked
  # out or not.
  if ($dir_revision_type == iREVI) {
    $query = $vob_db->select("
        SELECT id AS dir_entry_id, element_id AS file_id, file_name
        FROM   dir_entry
        WHERE  dir_id = $dir_revision_id;
    ");
  } elsif ($dir_revision_type == iCOUT) {
    # get the view private directory ID from the checkout object
    $dir_revision_id =
        $inode_cache[$vob_id]->{$dir_revision_id}->{view_private_id};
    $query = $vob_db->select("
        SELECT id AS dir_entry_id, file_id, file_name
        FROM   dir_private_entry
        WHERE  view_id = $view_id
           AND dir_id = $dir_revision_id;
    ");
  } else {
    warn "readdir_view(@_): revision returned " .
         "by select_revision_id() ($dir_revision_type) not REVI or COUT\n";
    return (NFSERR_IO);
  }

  # put the list of VOB entries in the directory into %vob_list
  while ($row = $query->fetchrow()) {
    if (get_inode_type($view_id, $row->{file_id}) != iELEM) {
      warn "directory $dir_revision_id " .
           "contains non-element object $row->{file_id}\n";
      next;
    }

    # checked out elements appear on top of view private files
    if (is_checked_out($view_id, $row->{file_id})) {

      # unless the element is a regular file...
      if ($inode_cache[$vob_id]->{$row->{file_id}}->{type} == eREG) {

        # ...and the config spec selects the CHECKEDOUT revision of
        # the element
        my($selected_object) =
            select_revision_id($view_id, $row->{file_id},
                               "$dir_dpath/$row->{file_name}");
        next if (get_inode_type($view_id, $selected_object) == iCOUT);
      }

      delete($view_list{$row->{file_name}});
    }

    # view private files appear on top of VOB elements
    # (except for checked-out files)
    next if $view_list{$row->{file_name}};

    $vob_list{$row->{file_name}} = $row;
  }

  my(@dir_list);
  my(%file_list) = (%vob_list, %view_list);

  # merge %vob_list and %view_list
  for my $dir_entry (sort { $file_list{$a}->{dir_entry_id} <=>
                            $file_list{$b}->{dir_entry_id} }
                          keys %file_list) {
    push(@dir_list, [
      $file_list{$dir_entry}->{file_id},
      $dir_entry,
      $file_list{$dir_entry}->{dir_entry_id},
    ]);
  }

  return (undef, @dir_list);
}

# given a directory of type dATAT, return a listing of the files in
# the directory
sub readdir_atat {
  my($view_id, $dir_inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($element_id) = get_element_id($view_id, $dir_inode_id);
  my(@dir_list) = ();
  my($query, $row);

  push(@dir_list, [
    $dir_inode_id,
    '.',
    1,
  ]);

  # We know that get_dir_type has already been called for this
  # $dir_inode_id
  push(@dir_list, [
    $inode_cache[$vob_id]->{$dir_inode_id}->{id},
    '..',
    2,
  ]);

  # Branches (ie. main)
  $query = $vob_db->select("
      SELECT element_branch.id, branch.name
      FROM   element_branch, branch
      WHERE  element_branch.element_id = $element_id
         AND element_branch.parent_revision_id IS NULL
         AND element_branch.branch_id = branch.id;
  ");
  return (NFSERR_IO) unless $row = $query->fetchrow();
  push(@dir_list, [
    $row->{id},
    $row->{name},
    3,
  ]);

  # Ensure only one main branch
  return (NFSERR_IO) if $query->fetchrow();

  # Labels
  $query = $vob_db->select("
      SELECT   label.ext_dir_entry_id, revision.id, label_type.name
      FROM     label, label_type, revision, element_branch
      WHERE    label.label_type_id = label_type.id
           AND NOT label_type.pbranch
           AND label.revision_id = revision.id
           AND revision.element_branch_id = element_branch.id
           AND element_branch.element_id = $element_id
      ORDER BY label.ext_dir_entry_id;
  ");
  while ($row = $query->fetchrow()) {
    push(@dir_list, [
      $row->{id},
      $row->{name},
      $row->{ext_dir_entry_id},
    ]);
  }

  return (undef, @dir_list);
}

# given a directory of type dELBR, return a listing of the files in
# the directory
sub readdir_elbr {
  my($view_id, $dir_inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my(@dir_list) = ();
  my($query, $row, $last_row);

  push(@dir_list, [
    $dir_inode_id,
    '.',
    1,
  ]);

  if (!defined($inode_cache[$vob_id]->{$dir_inode_id}->{parent_revision_id})) {
    my($element_id) = get_element_id($view_id, $dir_inode_id);
    push(@dir_list, [
      $inode_cache[$vob_id]->{$element_id}->{id2},
      '..',
      2,
    ]);
  } else {
    my($parent_revision_id) =
        $inode_cache[$vob_id]->{$dir_inode_id}->{parent_revision_id};

    # make sure that inode_cache has $parent_revision_id in it
    get_inode_type($view_id, $parent_revision_id);

    push(@dir_list, [
      $inode_cache[$vob_id]->{$parent_revision_id}->{element_branch_id},
      '..',
      2,
    ]);
  }

  # Revision numbers
  $query = $vob_db->select("
      SELECT   id, num, ext_dir_entry_id
      FROM     revision
      WHERE    element_branch_id = $dir_inode_id;
  ");
  while ($row = $query->fetchrow()) {
    push(@dir_list, [
      $row->{id},
      $row->{num},
      $row->{ext_dir_entry_id},
    ]);
    $last_row = $row;
  }
  if (!$last_row) {
    warn "readdir_elbr(@_): no revisions found\n";
  } else {
    push(@dir_list, [
      $last_row->{id},
      'LATEST',
      3,
    ]);
  }

  # Checkouts
  $query = $vob_db->select("
      SELECT checkout.id, checkout.view_id,
             ext_dir_entry_id_1, ext_dir_entry_id_2
      FROM   checkout, revision
      WHERE  checkout.revision_id = revision.id
         AND revision.element_branch_id = $dir_inode_id;
  ");
  while ($row = $query->fetchrow()) {
    if ($row->{view_id} == $view_id) {
      push(@dir_list, [
        $row->{id},
        "CHECKEDOUT",
        $row->{ext_dir_entry_id_1},
      ]);
    }
    push(@dir_list, [
      $row->{id},
      "CHECKEDOUT.$row->{id}",
      $row->{ext_dir_entry_id_2},
    ]);
  }

  # Labels
  $query = $vob_db->select("
      SELECT revision.id, label_type.name, label.ext_dir_entry_id
      FROM   label, label_type, revision
      WHERE  label.label_type_id = label_type.id
         AND label.revision_id = revision.id
         AND revision.element_branch_id = $dir_inode_id
  ");
  while ($row = $query->fetchrow()) {
    push(@dir_list, [
      $row->{id},
      $row->{name},
      $row->{ext_dir_entry_id},
    ]);
  }

  # Branches
  $query = $vob_db->select("
      SELECT element_branch.id, branch.name, element_branch.ext_dir_entry_id
      FROM   element_branch, branch, revision
      WHERE  element_branch.branch_id = branch.id
         AND element_branch.parent_revision_id = revision.id
         AND revision.element_branch_id = $dir_inode_id;
  ");
  while ($row = $query->fetchrow()) {
    push(@dir_list, [
      $row->{id},
      $row->{name},
      $row->{ext_dir_entry_id},
    ]);
  }

  # Sort by dir_entry_id
  @dir_list = sort { $a->[2] <=> $b->[2] } @dir_list;

  return (undef, @dir_list);
}

# given a directory of type dREVI, return a listing of the files in
# the directory
sub readdir_revi {
  my($view_id, $dir_inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my(@dir_list) = ();
  my($query, $row);

  $query = $vob_db->select("
      SELECT   id AS dir_entry_id, element_id, file_name
      FROM     dir_entry
      WHERE    dir_id = $dir_inode_id
      ORDER BY id;
  ");
  while ($row = $query->fetchrow()) {
    if (get_inode_type($view_id, $row->{element_id}) != iELEM) {
      warn "readdir_revi(@_): " .
           "get_inode_type($view_id, $row->{element_id}) != iELEM";
      next;
    }

    # If it's not '.' or '..', add the file as-is.
    # For '.'/'..' we need to use a different inode number.
    push(@dir_list, [
      $row->{file_name} eq '.'
        ? $dir_inode_id
        : $row->{file_name} eq '..'
        ? $inode_cache[$vob_id]->{$dir_inode_id}->{element_branch_id}
        : $inode_cache[$vob_id]->{$row->{element_id}}->{id2},
      $row->{file_name},
      $row->{dir_entry_id},
    ]);
  }

  return (undef, @dir_list);
}

# given a directory of type dVUPR, return a listing of the files in
# the directory
sub readdir_vupr {
  my($view_id, $dir_inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my(@dir_list) = ();
  my($query, $row);

  $query = $vob_db->select("
      SELECT   id AS dir_entry_id, file_id, file_name
      FROM     dir_private_entry
      WHERE    view_id = $view_id
           AND dir_id = $dir_inode_id
      ORDER BY id;
  ");
  while ($row = $query->fetchrow()) {
    push(@dir_list, [
      $row->{file_id},
      $row->{file_name},
      $row->{dir_entry_id},
    ]);
  }

  return (undef, @dir_list);
}

# given a directory of type dCOUT, return a listing of the files in
# the directory
sub readdir_cout {
  my($view_id, $dir_inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($vupr_dir_inode_id);
  my(@dir_list) = ();
  my($query, $row);

  $vupr_dir_inode_id =
      $inode_cache[$vob_id]->{$dir_inode_id}->{view_private_id};

  # get the element_branch_id for the directory given -- this is the
  # inode number for '..' in the directory we are looking at
  # XXX should be able to figure this out w/o a db query
  $row = $vob_db->select("
      SELECT revision.element_branch_id
      FROM   revision, checkout
      WHERE  revision.id = checkout.revision_id
         AND checkout.id = $dir_inode_id;
  ")->fetchrow();
  if (!$row) {
    warn "couldn't get element_branch_id for $dir_inode_id\n";
    return (NFSERR_IO);
  }
  my($element_branch_id) = $row->{element_branch_id};

  $query = $vob_db->select("
      SELECT   id AS dir_entry_id, file_id, file_name
      FROM     dir_private_entry
      WHERE    dir_id = $vupr_dir_inode_id
      ORDER BY id;
  ");
  while ($row = $query->fetchrow()) {
    if (get_inode_type($view_id, $row->{file_id}) != iELEM) {
      warn "readdir_cout(@_): " .
           "get_inode_type($view_id, $row->{file_id}) != iELEM";
      next;
    }

    # If it's not '.' or '..', add the file as-is.
    # For '.'/'..' we need to use a different inode number.
    push(@dir_list, [
      $row->{file_name} eq '.'
        ? $dir_inode_id
        : $row->{file_name} eq '..'
        ? $element_branch_id
        : $inode_cache[$vob_id]->{$row->{file_id}}->{id2},
      $row->{file_name},
      $row->{dir_entry_id},
    ]);
  }

  return (undef, @dir_list);
}

# view_readdir returns a list containing:
# - error number
# - eof flag
# - directory entries
#
# Each directory entry is an array ref with the elements containing:
# - inode number
# - file name
# - cookie
sub view_readdir {
  my($view_id, $dir_inode_id, $dir_dpath, $cookie, $count) = @_;
  my($dir_type);
  my($error, @dir_list, $eof);

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if (!$dir_type) {
    warn "nfs_readdir called with a non-directory " .
         "($view_id, $dir_inode_id)\n";
    return NFSERR_NOTDIR;
  }

  # each of these is expected to return a list sorted by the cookie
  if ($dir_type == dVIEW) {
    ($error, @dir_list) = readdir_view($view_id, $dir_inode_id, $dir_dpath);
  } elsif ($dir_type == dATAT) {
    ($error, @dir_list) = readdir_atat($view_id, $dir_inode_id);
  } elsif ($dir_type == dELBR) {
    ($error, @dir_list) = readdir_elbr($view_id, $dir_inode_id);
  } elsif ($dir_type == dREVI) {
    ($error, @dir_list) = readdir_revi($view_id, $dir_inode_id);
  } elsif ($dir_type == dVUPR) {
    ($error, @dir_list) = readdir_vupr($view_id, $dir_inode_id);
  } elsif ($dir_type == dCOUT) {
    ($error, @dir_list) = readdir_cout($view_id, $dir_inode_id);
  } else {
    die "readdir($view_id, $dir_inode_id): dir_type is $dir_type\n";
  }

  return ($error) if $error;

  # get rid of any dirlist entries where its cookie is less than the
  # given cookie
  shift @dir_list while @dir_list and $dir_list[0]->[2] <= $cookie;

  $eof = @dir_list <= $count;

  return (undef, $eof, $eof ? @dir_list : @dir_list[0..$count - 1]);
}

1;
