/* Katie
 *
 * Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

/* obj_id_seq for objects
 * obj IDs are reused for inode numbers by element, element_branch, and
 * revision
 */

CREATE SEQUENCE obj_id_seq;
GRANT ALL ON obj_id_seq TO katie;

/* directory entry IDs for objects that appear in view or view-private
 * directories
 */
CREATE SEQUENCE dir_entry_id_seq;
GRANT ALL ON dir_entry_id_seq TO katie;

/* extended directory entry IDs for objects that appear in the
 * version-extended space
 */
CREATE SEQUENCE ext_dir_entry_id_seq START 5;
GRANT ALL ON ext_dir_entry_id_seq TO katie;

/* Elements
 * id is the inode that the view sees
 * id2 is the inode for the @@ version of the element
 */
CREATE TABLE element (
  id    INT PRIMARY KEY DEFAULT nextval('obj_id_seq'),
  id2   INT NOT NULL DEFAULT nextval('obj_id_seq'),
  type  INT NOT NULL,
  mode  INT NOT NULL,
  uid   INT NOT NULL,
  gid   INT NOT NULL,
  mtime INT NOT NULL DEFAULT date_part('epoch', now())
);
GRANT ALL ON element TO katie;

/* Branches */
CREATE TABLE branch (
  id      INT PRIMARY KEY DEFAULT nextval('obj_id_seq'),
  name    VARCHAR(255) NOT NULL,
  uid     INT NOT NULL,
  gid     INT NOT NULL,
  pbranch BOOLEAN NOT NULL DEFAULT FALSE
);
GRANT ALL ON branch TO katie;

/* Branches on an element */
CREATE TABLE element_branch (
  id                 INT PRIMARY KEY DEFAULT nextval('obj_id_seq'),
  ext_dir_entry_id   INT NOT NULL DEFAULT nextval('ext_dir_entry_id_seq'),
  element_id         INT NOT NULL,
  branch_id          INT NOT NULL,
  parent_revision_id INT
);
GRANT ALL ON element_branch TO katie;

/* Every revision of every element */
CREATE TABLE revision (
  id                INT PRIMARY KEY DEFAULT nextval('obj_id_seq'),
  ext_dir_entry_id  INT NOT NULL DEFAULT nextval('ext_dir_entry_id_seq'),
  element_branch_id INT NOT NULL,
  num               INT NOT NULL,
  mtime             INT NOT NULL DEFAULT date_part('epoch', now()),
  size              INT NOT NULL
);
GRANT ALL ON revision TO katie;

/* Checked-out revisions */
CREATE TABLE checkout (
  id                 INT PRIMARY KEY DEFAULT nextval('obj_id_seq'),
  ext_dir_entry_id_1 INT NOT NULL DEFAULT nextval('ext_dir_entry_id_seq'),
  ext_dir_entry_id_2 INT NOT NULL DEFAULT nextval('ext_dir_entry_id_seq'),
  event_id           INT NOT NULL,
  element_id         INT NOT NULL,
  revision_id        INT NOT NULL,
  view_id            INT NOT NULL,
  reserved           BOOLEAN DEFAULT TRUE,
  mtime              INT NOT NULL DEFAULT date_part('epoch', now()),

  /* for directories only, the view_private_id is kept */
  view_private_id INT
);
GRANT ALL ON checkout TO katie;
CREATE UNIQUE INDEX checkout_element_view ON checkout (element_id, view_id);

/* View-private inode numbers are allocated as:
 *   2^31 - 1 - nextval('private_id_seq')
 */
CREATE SEQUENCE private_id_seq;
GRANT ALL ON private_id_seq TO katie;

/* Index to view-private files */
CREATE TABLE view_private (
  id         INT PRIMARY KEY DEFAULT 2147483647 - nextval('private_id_seq'),
  link_count INT NOT NULL DEFAULT 1,
  type       INT NOT NULL,
  mode       INT NOT NULL,
  uid        INT NOT NULL,
  gid        INT NOT NULL
);
GRANT ALL ON view_private TO katie;

CREATE TABLE dir_entry (
  id           INT PRIMARY KEY DEFAULT nextval('dir_entry_id_seq'),
  dir_id       INT NOT NULL,
  element_id   INT NOT NULL,
  file_name    VARCHAR(255) NOT NULL
);
GRANT ALL ON dir_entry TO katie;
CREATE INDEX dir_entry_dir_id ON dir_entry (dir_id);

CREATE TABLE dir_private_entry (
  id           INT PRIMARY KEY DEFAULT nextval('dir_entry_id_seq'),
  view_id      INT NOT NULL,
  dir_id       INT NOT NULL,
  file_id      INT NOT NULL,
  file_name    VARCHAR(255) NOT NULL
);
GRANT ALL ON dir_private_entry TO katie;
CREATE INDEX dir_private_entry_view_id ON dir_private_entry (view_id);
CREATE INDEX dir_private_entry_dir_id ON dir_private_entry (dir_id);

CREATE TABLE dir_mtime (
  view_id INT NOT NULL,
  dir_id  INT NOT NULL,
  mtime   INT NOT NULL DEFAULT date_part('epoch', now())
);
GRANT ALL ON dir_mtime TO katie;
CREATE UNIQUE INDEX dir_mtime_view_dir ON dir_mtime (view_id, dir_id);

/* Labels */
CREATE TABLE label_type (
  id      INT PRIMARY KEY DEFAULT nextval('obj_id_seq'),
  name    VARCHAR(255) NOT NULL UNIQUE,
  uid     INT NOT NULL,
  gid     INT NOT NULL,
  pbranch BOOLEAN NOT NULL DEFAULT FALSE
);
GRANT ALL ON label_type TO katie;

CREATE TABLE label (
  ext_dir_entry_id INT PRIMARY KEY DEFAULT nextval('ext_dir_entry_id_seq'),
  label_type_id    INT NOT NULL,
  revision_id      INT NOT NULL
);
GRANT ALL ON label TO katie;
CREATE UNIQUE INDEX label_revision_type ON label (revision_id, label_type_id);
CREATE INDEX label_label_type ON label (label_type_id);

/* Attributes */
CREATE TABLE attribute_type (
  id     INT PRIMARY KEY DEFAULT nextval('obj_id_seq'),
  name   VARCHAR(255) NOT NULL UNIQUE,

  /* per element, per branch, per version */
  per    CHAR CHECK (per IN ('E', 'B', 'V')),

  /* int, real, time, string */
  attype CHAR NOT NULL CHECK (attype IN ('I', 'R', 'T', 'S'))
);
GRANT ALL ON attribute_type TO katie;

CREATE TABLE attribute (
  attribute_type_id INT NOT NULL,
  object_id         INT NOT NULL,
  value             TEXT NOT NULL
);
GRANT ALL ON attribute TO katie;
CREATE UNIQUE INDEX attribute_obj_attr
    ON attribute (object_id, attribute_type_id);
CREATE INDEX attribute_attr_type ON attribute (attribute_type_id);

/* Config Specs */
CREATE TABLE config_spec (
  view_id INT PRIMARY KEY,
  spec    TEXT NOT NULL,
  mtime   INT NOT NULL DEFAULT date_part('epoch', now())
);
GRANT ALL ON config_spec TO katie;

/* Events */
CREATE SEQUENCE event_id_seq;
GRANT ALL ON event_id_seq TO katie;

CREATE TABLE event (
  id         INT NOT NULL DEFAULT nextval('event_id_seq'),
  object_id  INT NOT NULL,
  event_type VARCHAR(20) NOT NULL
             CHECK (event_type IN ('checkout', 'create', 'modify')),
  uid        INT NOT NULL,
  time       INT NOT NULL DEFAULT date_part('epoch', now()),
  comment    TEXT NOT NULL DEFAULT ''
);
GRANT ALL ON event TO katie;

/*
INSERT INTO type (descr) VALUES ('dir');
INSERT INTO type (descr) VALUES ('symlink');
INSERT INTO type (descr) VALUES ('file');
*/
